import os
import pickle
import logging

import torch
import numpy as np
import pandas as pd
from recbole.data.interaction import Interaction
from recbole.data.dataset import Dataset as RecboleDataset


class Dataset(RecboleDataset):
    SPLITS = ['train', 'validation', 'test']

    def __init__(self, config):
        if 'LRS' not in config['eval_args']['split']:
            raise ValueError("Perturbed graph can be used only when splits are loaded.")

        super(Dataset, self).__init__(config)

    def _discretization(self):
        if self.config['numerical_features'] is None:
            self.config['numerical_features'] = []
        super(Dataset, self)._discretization()

    def _load_data_split(self, split):
        filename = os.path.join(self.dataset_path, f'{self.dataset_name}.{split}')
        if not os.path.isfile(filename):
            if split in ['train', 'test']:
                raise NotImplementedError(f'The splitting method "LRS" needs at least train and test.')
            return None

        with open(filename, 'rb') as split_file:
            split_data: dict = pickle.load(split_file)

        return split_data

    def split_by_loaded_splits(self):
        next_df = []
        for split in self.SPLITS:
            split_data = self._load_data_split(split)

            if split_data is None:
                next_df.append(None)
            else:
                split_keys = list(split_data.keys())
                if self.uid_field not in split_keys or self.iid_field not in split_keys or len(split_keys) != 2:
                    raise ValueError(f'The splitting grouping method "LRS" should contain only the fields '
                                     f'`{self.uid_field}` and `{self.iid_field}`.')

                for field in [self.uid_field, self.iid_field]:
                    data = split_data[field]
                    data = data.numpy() if isinstance(data, torch.Tensor) else data
                    split_data[field] = torch.LongTensor([self.field2token_id[field][val] for val in data.astype(str)])

                next_df.append(Interaction(split_data))

        self._drop_unused_col()
        next_ds = [self.copy(_) if _ is not None else None for _ in next_df]
        return next_ds

    def build(self):
        """Processing dataset according to evaluation setting, including Group, Order and Split.
        See :class:`~recbole.config.eval_setting.EvalSetting` for details.

        Returns:
            list: List of built :class:`Dataset`.
        """
        self._change_feat_format()

        if self.benchmark_filename_list is not None:
            self._drop_unused_col()
            cumsum = list(np.cumsum(self.file_size_list))
            datasets = [
                self.copy(self.inter_feat[start:end])
                for start, end in zip([0] + cumsum[:-1], cumsum)
            ]
            return datasets

        # ordering
        ordering_args = self.config["eval_args"]["order"]
        if ordering_args == "RO":
            self.shuffle()
        elif ordering_args == "TO":
            self.sort(by=self.time_field)
        else:
            raise NotImplementedError(
                f"The ordering_method [{ordering_args}] has not been implemented."
            )

        # splitting & grouping
        split_args = self.config["eval_args"]["split"]
        if split_args is None:
            raise ValueError("The split_args in eval_args should not be None.")
        if not isinstance(split_args, dict):
            raise ValueError(f"The split_args [{split_args}] should be a dict.")

        split_mode = list(split_args.keys())[0]
        assert len(split_args.keys()) == 1
        group_by = self.config["eval_args"]["group_by"]
        if split_mode == "RS":
            if not isinstance(split_args["RS"], list):
                raise ValueError(f'The value of "RS" [{split_args}] should be a list.')
            if group_by is None or group_by.lower() == "none":
                datasets = self.split_by_ratio(split_args["RS"], group_by=None)
            elif group_by == "user":
                datasets = self.split_by_ratio(
                    split_args["RS"], group_by=self.uid_field
                )
            else:
                raise NotImplementedError(
                    f"The grouping method [{group_by}] has not been implemented."
                )
        elif split_mode == "LS":
            datasets = self.leave_one_out(
                group_by=self.uid_field, leave_one_mode=split_args["LS"]
            )
        elif split_mode == "LRS":  # Load Ready Splits
            datasets = self.split_by_loaded_splits()
        else:
            raise NotImplementedError(
                f"The splitting_method [{split_mode}] has not been implemented."
            )

        return datasets



# Based on https://github.com/RexYing/gnn-model-explainer/blob/master/explainer/explain.py

import os
import time
import math
from logging import getLogger

import wandb
import torch
import numpy as np
import pandas as pd
import seaborn as sns
import scipy.signal as sp_signal
import matplotlib.pyplot as plt
from tqdm import tqdm
from recbole.evaluator import Evaluator
from recbole.data.interaction import Interaction
from recbole.utils import set_color
import torch.distributed as torch_dist
import torch.multiprocessing as torch_mp
import torch.nn.parallel as torch_parallel  # import DistributedDataParallel as DDP

import gnnuers.utils as utils
import gnnuers.models as exp_models
import gnnuers.evaluation as eval_utils
from gnnuers.utils.early_stopping import EarlyStopping
from gnnuers.losses import get_ranking_loss, get_loss_from_exp_metric

from . import utils as exp_utils
from . import metrics as exp_metrics


class Explainer:

    def __init__(self, config, dataset, rec_data, model, dist="damerau_levenshtein", **kwargs):
        super(Explainer, self).__init__()
        self.config = config

        # self.cf_model = None
        self.model = model
        self.model.eval()
        self._cf_model = None

        self.dataset = dataset
        self.rec_data = rec_data
        self._pred_as_rec = config['exp_rec_data'] == 'rec'
        self._test_history_matrix = None

        self.cf_optimizer = None
        self.mini_batch_descent = config['mini_batch_descent']

        self.beta = config['cf_beta']
        self.device = config['device']
        self.unique_graph_dist_loss = config['save_unique_graph_dist_loss']
        self.old_graph_dist = 0

        self.tot_item_num = dataset.item_num
        self.item_tensor = dataset.get_item_feature().to(model.device)
        self.test_batch_size = self.tot_item_num

        if dist == "set":
            self.dist = lambda topk_idx, cf_topk_idx: len(topk_idx) - (len(set(topk_idx) & set(cf_topk_idx)))
        elif dist == "damerau_levenshtein":
            self.dist = utils.damerau_levenshtein_distance

        self.topk_args = None
        self.model_scores, self.model_scores_topk, self.model_topk_idx = None, None, None

        self.verbose = kwargs.get("verbose", False)
        self.logger = getLogger()

        self.user_batch_exp = config['user_batch_exp']

        self.eval_metric = config['eval_metric'] or 'ndcg'
        self.evaluator = Evaluator(config)
        self.exp_metric = config['exp_metric'] or 'consumer_DP_across_random_samples'

        self._metric_loss = get_ranking_loss(config['metric_loss'] or 'ndcg')
        self._exp_loss = get_loss_from_exp_metric(self.exp_metric)

        self.sensitive_attribute = config['sensitive_attribute']
        attr_map = dataset.field2id_token[self.sensitive_attribute]
        self.f_idx, self.m_idx = (attr_map == 'F').nonzero()[0][0], (attr_map == 'M').nonzero()[0][0]

        self.results = None
        self.adv_group, self.disadv_group = None, None
        self.only_adv_group = config['only_adv_group']

        self.earlys = EarlyStopping(
            config['early_stopping']['patience'],
            config['early_stopping']['ignore'],
            method=config['early_stopping']['method'],
            fn=config['early_stopping']['mode'],
            delta=config['early_stopping']['delta']
        )
        self.earlys_check_value = config['early_stopping']['check_value']

        self.previous_loss_value = config['previous_loss_value']
        self.previous_batch_LR_scaling = config['previous_batch_LR_scaling']
        self.lr_scaler = None

        # Init policies
        self.gradient_deactivation_constraint = config['explainer_policies']['gradient_deactivation_constraint'] or False
        self.increase_disparity = config['explainer_policies']['increase_disparity']
        self.group_deletion_constraint = config['explainer_policies']['group_deletion_constraint']
        self.random_perturbation = config['explainer_policies']['random_perturbation']
        self.neighborhood_perturbation = config['explainer_policies']['neighborhood_perturbation']
        self.users_zero_constraint = config['explainer_policies']['users_zero_constraint']
        self.users_zero_constraint_value = config['users_zero_constraint_value'] or 0
        self.users_low_degree = config['explainer_policies']['users_low_degree']
        self.users_low_degree_ratio = config['users_low_degree_ratio'] or 0.2
        self.items_preference_constraint = config['explainer_policies']['items_preference_constraint']
        self.items_preference_constraint_ratio = config['items_preference_constraint_ratio'] or 0.3
        self.users_furthest_constraint = config['explainer_policies']['users_furthest_constraint']
        self.users_furthest_constraint_ratio = config['users_furthest_constraint_ratio'] or 0.3
        self.sparse_users_constraint = config['explainer_policies']['sparse_users_constraint']
        self.sparse_users_constraint_ratio = config['sparse_users_constraint_ratio'] or 0.3
        self.niche_items_constraint = config['explainer_policies']['niche_items_constraint']
        self.niche_items_constraint_ratio = config['niche_items_constraint_ratio'] or 0.3

        self.coverage_min_relevant_items = config['coverage_min_relevant_items'] or 0
        self.coverage_loss_only_relevant = config['coverage_loss_only_relevant'] or True

        self.item_discriminative_attribute = config['item_discriminative_attribute'] or 'exposure'
        self.item_discriminative_ratio = config['short_head_item_discriminative_ratio'] or 0.2  # SH / LT Explainable Fairness in Recommendation
        self.item_discriminative_groups_distrib = [1, 1 / self.item_discriminative_ratio - 1]
        self.item_discriminative_map = ['[PAD]', 'SH', 'LT']  # SH: Short Head, LT: Long Tail

        self._exp_loss_args = {
            'consumer_dp': [self.sensitive_attribute],
            'consumer_dp_across_random_samples': [self.sensitive_attribute],
            'provider_dp': [self.item_discriminative_attribute],
            'uc': [self.coverage_min_relevant_items]
        }

        self.ckpt_loading_path = None

    @property
    def cf_model(self):
        if self._cf_model is None:
            print("Counterfactual Model Explainer is not initialized yet. Execute 'explain' to initialize it.")
        else:
            return self._cf_model

    @cf_model.setter
    def cf_model(self, value):
        self._cf_model = value

    def initialize_cf_model(self, **kwargs):
        # Instantiate CF model class, load weights from original model
        self.cf_model = getattr(
            exp_models,
            f"{self.model.__class__.__name__}Perturbed"
        )(self.config, self.dataset, **kwargs).to(self.model.device)
        # self.parallel_cf_model = torch_parallel.DistributedDataParallel(self.cf_model)
        # self.parallel_cf_model = torch_parallel.DataParallel(self.cf_model)
        # for attr in ['device', 'full_sort_predict']:
        #     setattr(self.parallel_cf_model, attr, getattr(self.cf_model, attr))
        #
        # self.cf_model = self.parallel_cf_model
        self.cf_model.load_state_dict(self.model.state_dict(), strict=False)

        # Freeze weights from original model in cf_model
        for name, param in self.cf_model.named_parameters():
            if name != "P_symm":
                param.requires_grad = False

    def initialize_optimizer(self):
        if self.cf_model is None:
            raise AttributeError("The counterfactual model `cf_model` should be initialized before the optimizer.")

        lr = self.config['cf_learning_rate']
        momentum = self.config["momentum"] or 0.0
        sgd_kwargs = {'momentum': momentum, 'nesterov': True if momentum > 0 else False}
        if self.config["cf_optimizer"] == "SGD":
            self.cf_optimizer = torch.optim.SGD(self.cf_model.parameters(), lr=lr, **sgd_kwargs)
        elif self.config["cf_optimizer"] == "Adadelta":
            self.cf_optimizer = torch.optim.Adadelta(self.cf_model.parameters(), lr=lr)
        elif self.config["cf_optimizer"] == "Adagrad":
            self.cf_optimizer = torch.optim.Adagrad(self.cf_model.parameters(), lr=lr)
        elif self.config["cf_optimizer"] == "AdamW":
            self.cf_optimizer = torch.optim.AdamW(self.cf_model.parameters(), lr=lr)
        elif self.config["cf_optimizer"] == "Adam":
            self.cf_optimizer = torch.optim.Adam(self.cf_model.parameters(), lr=lr)
        elif self.config["cf_optimizer"] == "RMSprop":
            self.cf_optimizer = torch.optim.RMSprop(self.cf_model.parameters(), lr=lr)
        else:
            raise NotImplementedError("CF Optimizer not implemented")

    def set_checkpoint_path(self, path):
        self.ckpt_loading_path = path

    def _resume_checkpoint(self):
        ckpt = torch.load(self.ckpt_loading_path)
        epoch = ckpt['starting_epoch']
        exp_losses = ckpt['exp_losses'] if 'exp_losses' in ckpt else ckpt['fair_losses']
        self.earlys = ckpt['early_stopping']
        best_cf_example = ckpt['best_cf_example']
        self.cf_model.load_cf_state_dict(ckpt)

        return exp_losses, epoch, best_cf_example

    def _save_checkpoint(self, epoch, exp_losses, best_cf_example):
        cf_state_dict = self.cf_model.cf_state_dict()
        ckpt = {
            'starting_epoch': epoch,
            'exp_losses': exp_losses,
            'early_stopping': self.earlys,
            'best_cf_example': best_cf_example,
            **cf_state_dict
        }
        torch.save(ckpt, self.ckpt_loading_path)

    def compute_model_predictions(self, scores_args, topk):
        """
        Compute the predictions of the original model without perturbation
        :param topk: integer of topk items
        :param scores_args: arguments needed by recbole to compute scores
        :return:
        """
        self.topk_args = {'topk': topk}

        self.model_scores = self.get_scores(self.model, *scores_args, pred=None)

        # topk_idx contains the ids of the topk items
        self.model_scores_topk, self.model_topk_idx = self.get_top_k(self.model_scores, **self.topk_args)

    def _get_scores_args(self, batched_data, dataset):
        dset_batch_data = Explainer.prepare_batched_data(batched_data, dataset)
        return [dset_batch_data, self.tot_item_num, self.test_batch_size, self.item_tensor]

    def _get_model_score_data(self, batched_data, dataset, topk):
        dset_scores_args = self._get_scores_args(batched_data, dataset)
        self.compute_model_predictions(dset_scores_args, topk)
        dset_model_topk = self.model_topk_idx.detach().cpu().numpy()

        return dset_scores_args, dset_model_topk

    def _get_no_grad_pred_model_score_data(self, scores_args, model_topk=None, compute_dist=False):
        self.cf_model.eval()
        # When recommendations are generated passing test set the items in train and validation are considered watched
        with torch.no_grad():
            cf_scores_pred = self.get_scores(self.cf_model, *scores_args, pred=True)
            _, cf_topk_pred_idx = self.get_top_k(cf_scores_pred, **self.topk_args)
        cf_topk_pred_idx = cf_topk_pred_idx.detach().cpu().numpy()

        cf_dist = None
        if compute_dist and model_topk is not None:
            cf_dist = [
                self.dist(_pred, _topk_idx) for _pred, _topk_idx in zip(cf_topk_pred_idx, model_topk)
            ]

        return cf_topk_pred_idx, cf_dist

    def logging_exp_per_group(self, new_example, model_topk):
        em_str = self.eval_metric.upper()

        pref_data = pd.DataFrame(zip(*new_example[:2], model_topk), columns=['user_id', 'cf_topk_pred', 'topk_pred'])
        orig_res = self.compute_eval_metric(self.rec_data.dataset, pref_data, 'topk_pred')
        cf_res = self.compute_eval_metric(self.rec_data.dataset, pref_data, 'cf_topk_pred')

        user_feat = Interaction({k: v[torch.tensor(new_example[0])] for k, v in self.dataset.user_feat.interaction.items()})

        m_users = user_feat[self.sensitive_attribute] == self.m_idx
        f_users = user_feat[self.sensitive_attribute] == self.f_idx

        orig_f, orig_m = np.mean(orig_res[f_users, -1]), np.mean(orig_res[m_users, -1])
        cf_f, cf_m = np.mean(cf_res[f_users, -1]), np.mean(cf_res[m_users, -1])
        self.logger.info(f"Original => {em_str} F: {orig_f}, {em_str} M: {orig_m}, Diff: {np.abs(orig_f - orig_m)} \n"
                         f"CF       => {em_str} F: {cf_f}, {em_str} M: {cf_m}, Diff: {np.abs(cf_f - cf_m)}")

    def log_epoch(self, initial_time, epoch, *losses, **verbose_kws):
        loss_total, exp_loss, loss_graph_dist, orig_loss_graph_dist = losses
        self.logger.info(f"{self.cf_model.__class__.__name__} " +
                         f"Explain duration: {time.strftime('%H:%M:%S', time.gmtime(time.time() - initial_time))}, " +
                         # 'User id: {}, '.format(str(users_ids)) +
                         'Epoch: {}, '.format(epoch + 1) +
                         'loss: {:.4f}, '.format(loss_total.item()) +
                         'exp loss: {:.4f}, '.format(exp_loss) +
                         'graph loss: {:.4f}, '.format(loss_graph_dist.item()) +
                         'perturbed edges: {:.4f}, '.format(int(orig_loss_graph_dist.item())))
        if self.verbose:
            self.logger.info('Orig output: {}\n'.format(self.model_scores) +
                             # 'Output: {}\n'.format(verbose_kws.get('cf_scores', None)) +
                             # 'Output nondiff: {}\n'.format(verbose_kws.get('cf_scores_pred', None)) +
                             '{:20}: {},\n {:20}: {},\n {:20}: {}\n'.format(
                                 'orig pred', self.model_topk_idx,
                                 'new pred', verbose_kws.get('cf_topk_idx', None),
                                 'new pred nondiff', verbose_kws.get('cf_topk_pred_idx', None))
                             )
        self.logger.info(" ")

    def update_best_cf_example(self,
                               best_cf_example,
                               new_example,
                               loss_total,
                               best_loss,
                               model_topk=None,
                               force_update=False):
        """
        Updates the explanations with new explanation (if not None) depending on new loss value
        :param best_cf_example:
        :param new_example:
        :param loss_total:
        :param best_loss:
        :return:
        """
        if force_update or (new_example is not None and (abs(loss_total) < best_loss or self.unique_graph_dist_loss)):
            if self.unique_graph_dist_loss and len(best_cf_example) > 0:
                self.old_graph_dist = best_cf_example[-1][1]
                new_graph_dist = new_example[1]
                if not force_update and not (self.old_graph_dist != new_graph_dist):
                    return best_loss

            best_cf_example.append(new_example)
            self.old_graph_dist = new_example[1]

            if self.verbose and model_topk is not None:
                self.logging_exp_per_group(new_example, model_topk)

            if not self.unique_graph_dist_loss:
                return abs(loss_total)
        return best_loss

    def update_new_example(self,
                           new_example,
                           detached_batched_data,
                           full_dataset,
                           train_data,
                           valid_data,
                           test_data,
                           test_model_topk,
                           rec_model_topk,
                           epoch_exp_loss):
        perturbed_edges = new_example[-2]

        try:
            if self.config['exp_rec_data'] == 'rec':
                raise NotImplementedErorr(
                    'exp metric evaluation with dataloaders with perturbed edges not implemented for `exp_rec_data` == `rec`'
                )

            try:
                pert_sets = utils.get_dataloader_with_perturbed_edges(
                    perturbed_edges, self.config, full_dataset, train_data, valid_data, test_data
                )
            except ValueError as e:
                if repr(e) == "ValueError('Some users have interacted with all items, " \
                              "which we can not sample negative items for them. " \
                              "Please set `user_inter_num_interval` to filter those users.')":
                    return False, None
                raise(e)
            # import copy
            # conf = copy.deepcopy(self.config)
            # dd = copy.deepcopy(full_dataset)

            # import pdb; pdb.set_trace()
            # conf['user_inter_num_interval'] = f"[0, {dd.item_num}]"
            # dd.config['user_inter_num_interval'] = f"[0, {dd.item_num}]"
            # dd._from_scratch()
            # dd._filter_by_inter_num()

            # pert_sets = utils.get_dataloader_with_perturbed_edges(
            #     perturbed_edges, conf, dd, train_data, valid_data, test_data
            # )
            # pert_sets = utils.get_dataloader_with_perturbed_edges(
            #     perturbed_edges, self.config, full_dataset, train_data, valid_data, test_data
            # )
            pert_sets_dict = dict(zip(['train', 'valid', 'test'], pert_sets))

            test_scores_args = self._get_scores_args(detached_batched_data, pert_sets_dict['test'])
            rec_scores_args = self._get_scores_args(detached_batched_data, pert_sets_dict[self.config['exp_rec_data']])

            test_cf_topk_pred_idx, test_cf_dist = self._get_no_grad_pred_model_score_data(
                test_scores_args, model_topk=test_model_topk, compute_dist=True
            )
            rec_cf_topk_pred_idx, rec_cf_dist = self._get_no_grad_pred_model_score_data(
                rec_scores_args, model_topk=rec_model_topk, compute_dist=True
            )

            # new_example = [
            #     detached_batched_data,
            #     # rec_model_topk,
            #     # test_model_topk,
            #     rec_cf_topk_pred_idx,
            #     test_cf_topk_pred_idx,
            #     rec_cf_dist,
            #     test_cf_dist,
            #     *new_example[4:]
            # ]

            epoch_rec_exp_metric = self.compute_exp_metric(
                detached_batched_data,
                rec_cf_topk_pred_idx,
                pert_sets_dict[self.config['exp_rec_data']].dataset
            )

            epoch_test_exp_metric = self.compute_exp_metric(
                detached_batched_data,
                test_cf_topk_pred_idx,
                pert_sets_dict['test'].dataset
            )
        except TypeError as e:
            if self.earlys_check_value != 'exp_loss':
                raise NotImplementedError(
                    'must check how to solve the problem when certain users have no items in the history like after removing/adding edges'
                ) from e
            else:
                epoch_rec_exp_metric = None
                epoch_test_exp_metric = None

#         def pert_edges_mapper(pe, rec_dset):
#             return pe

#         test_pert_df, valid_pert_df = eval_utils.extract_metrics_from_perturbed_edges(
#             {(self.dataset.dataset_name, self.sensitive_attribute): perturbed_edges},
#             models=[self.model.__class__.__name__],
#             metrics=[self.eval_metric],
#             models_path=os.path.join(os.getcwd(), 'saved'),
#             remap=pert_edges_mapper
#         )

        new_example[utils.exp_col_index('exp_loss')] = epoch_exp_loss
        new_example[utils.exp_col_index('exp_metric')] = epoch_rec_exp_metric

        wandb.log({
            'loss': epoch_exp_loss,
            'rec_exp_metric': epoch_rec_exp_metric,
            'test_exp_metric': epoch_test_exp_metric,
            '# Del Edges': perturbed_edges.shape[1],
            'epoch': new_example[-1]
        })

        self.logger.info(str({
            'loss': epoch_exp_loss,
            'rec_exp_metric': epoch_rec_exp_metric,
            'test_exp_metric': epoch_test_exp_metric,
            '# Del Edges': perturbed_edges.shape[1],
            'epoch': new_example[-1]
        }))

        return new_example, epoch_rec_exp_metric

    @staticmethod
    def prepare_batched_data(batched_data, data, item_data=None):
        return utils.prepare_batched_data(batched_data, data, item_data=item_data)

    def get_iter_data(self, user_data):
        user_data = user_data.split(self.user_batch_exp)

        return (
            tqdm(
                user_data,
                total=len(user_data),
                ncols=100,
                desc=set_color(f"Explaining   ", 'pink'),
            )
        )

    def randperm2groups(self, batched_data):
        """
        At least 2 groups are represented in the batch following the distribution in the dataset.
        eps is used to select as an offset with respect to the fixed distribution. If a group has a 70% of distribution
        and the batch size is 32, then 22 +- (22 * eps) items are allocated for that group and the current batch
        :param batched_data:
        :return:
        """
        iter_data = []
        n_samples = batched_data.shape[0]
        n_batch = math.ceil(n_samples / self.user_batch_exp)

        attr = self.sensitive_attribute
        user_feat = self.dataset.user_feat[attr][batched_data]
        groups = user_feat.unique().numpy()

        masks = []
        for gr in groups:
            masks.append((user_feat == gr).numpy())
        masks = np.stack(masks)

        distrib = []
        for mask in masks:
            distrib.append(mask.nonzero()[0].shape[0] / batched_data.shape[0])

        for batch in range(n_batch):
            distrib = []
            for mask in masks:
                distrib.append(mask.nonzero()[0].shape[0] / n_samples)

            batch_len = min(n_samples, self.user_batch_exp)  # n_samples is lower than user_batch only for last batch
            batch_counter = batch_len
            batch_data = []
            for mask_i, mask_idx in enumerate(np.random.permutation(np.arange(masks.shape[0]))):
                if mask_i == (masks.shape[0] - 1):
                    n_mask_samples = batch_counter
                else:
                    if batch_counter < batch_len:
                        n_mask_samples = max(min(round(distrib[mask_idx] * batch_len), batch_counter), 1)
                    else:
                        n_mask_samples = max(min(round(distrib[mask_idx] * batch_len), batch_counter - 1), 1)
                mask_samples = np.random.permutation(masks[mask_idx].nonzero()[0])
                if batch != (n_batch - 1):
                    if mask_samples.shape[0] == n_mask_samples:
                        n_mask_samples = max(n_mask_samples - (n_batch - 1) - batch, 1)

                    mask_samples = mask_samples[:n_mask_samples]
                batch_data.append(batched_data[mask_samples])
                masks[mask_idx, mask_samples] = False  # affect groups where these users belong (e.g. gender and age group)
                batch_counter -= mask_samples.shape[0]
                n_samples -= mask_samples.shape[0]

                if batch_counter == 0:
                    break
            iter_data.append(torch.cat(batch_data))

        return iter_data

    def prepare_iter_batched_data(self, batched_data):
        if self._exp_loss.loss_type() == 'Consumer':
            if self.only_adv_group != "global":
                iter_data = self.randperm2groups(batched_data)
                # check if each batch has at least 2 groups
                while any(self.dataset.user_feat[self.sensitive_attribute][d].unique().shape[0] < 2 for d in iter_data):
                    iter_data = self.randperm2groups(batched_data)
            else:
                batched_attr_data = self.dataset.user_feat[self.sensitive_attribute][batched_data]
                iter_data = batched_data[batched_attr_data == self.adv_group].split(self.user_batch_exp)
        else:
            iter_data = batched_data[torch.randperm(batched_data.shape[0])].split(self.user_batch_exp)

        return iter_data

    def _prepare_test_history_matrix(self, test_data, topk=10):
        uids = test_data.dataset.user_feat.interaction[test_data.dataset.uid_field][1:]

        dset_scores_args, dset_model_topk = self._get_model_score_data(uids, test_data, topk)

        # add -1 as item ids for the padding user
        dset_model_topk = np.vstack((np.array([-1] * topk, dtype=dset_model_topk.dtype), dset_model_topk))
        self._test_history_matrix = dset_model_topk

        return dset_scores_args, dset_model_topk

    @staticmethod
    def _verbose_plot(exp_losses, epoch):
        if os.path.isfile(f'loss_trend_epoch{epoch}.png'):
            os.remove(f'loss_trend_epoch{epoch}.png')
        ax = sns.lineplot(
            x='epoch',
            y='exp loss',
            data=pd.DataFrame(zip(np.arange(1, epoch + 2), exp_losses), columns=['epoch', 'exp loss'])
        )
        if len(exp_losses) > 20:
            sns.lineplot(
                x=np.arange(1, epoch + 2),
                y=sp_signal.savgol_filter(exp_losses, window_length=len(exp_losses) // 2, polyorder=2),
                ax=ax
            )
        plt.savefig(f'loss_trend_epoch{epoch + 1}.png')
        plt.close()

    def _check_loss_trend_epoch_images(self):
        cwd_files = [f for f in os.listdir() if f.startswith('loss_trend_epoch')]
        if len(cwd_files) > 0 and os.path.isfile(cwd_files[0]) and cwd_files[0][-3:] == 'png':
            os.remove(cwd_files[0])

    def _check_policies(self, batched_data, rec_model_topk):
        filtered_users = None
        filtered_items = None
        test_model_topk, test_scores_args, rec_scores_args = [None] * 3
        if self.increase_disparity:
            batched_data, test_model_topk, test_scores_args,\
                rec_model_topk, rec_scores_args = self.increase_dataset_unfairness(
                    batched_data,
                    test_data,
                    rec_model_topk,
                    topk=topk
                )

            filtered_users = batched_data

        self.determine_adv_group(batched_data.detach().numpy(), rec_model_topk)

        if self.group_deletion_constraint and self.random_perturbation:
            raise NotImplementedError(
                'The policies `group_deletion_constraint` and `random_perturbation` cannot be both True'
            )
        elif self.group_deletion_constraint:
            if filtered_users is None:
                filtered_users = batched_data

            filtered_users = filtered_users[
                self.dataset.user_feat[self.sensitive_attribute][filtered_users] == self.adv_group
            ]

            if self.users_furthest_constraint:
                disadv_users = batched_data[
                    self.dataset.user_feat[self.sensitive_attribute][batched_data] == self.disadv_group
                ].numpy()

                igg = eval_utils.get_bipartite_igraph(self.dataset, remove_first_row_col=True)
                mean_dist = np.array(igg.distances(source=filtered_users, target=disadv_users)).mean(axis=1)
                furthest_users = np.argsort(mean_dist)

                filtered_users = filtered_users[
                    furthest_users[-int(self.users_furthest_constraint_ratio * furthest_users.shape[0]):]
                ]
        elif self.random_perturbation:
            # overwrites `increase_disparity` policy
            filtered_users = exp_models.PerturbedModel.RANDOM_POLICY

        if self.users_zero_constraint and self.random_perturbation:
            raise NotImplementedError(
                'The policies `users_zero_constraint` and `random_perturbation` cannot be both True'
            )
        elif self.users_zero_constraint:
            if filtered_users is None:
                filtered_users = batched_data

            pref_data = self._pref_data_sens_and_metric(batched_data.detach().numpy(), rec_model_topk)
            zero_users = pref_data.loc[(pref_data[self.eval_metric] <= self.users_zero_constraint_value), 'user_id']
            filtered_users = torch.from_numpy(np.intersect1d(zero_users.to_numpy(), filtered_users.numpy()))

        if self.users_low_degree and self.random_perturbation:
            raise NotImplementedError(
                'The policies `users_low_degree` and `random_perturbation` cannot be both True'
            )
        elif self.users_low_degree:
            if filtered_users is None:
                filtered_users = batched_data

            _, _, hist_len = self.dataset.history_item_matrix()
            hist_len = hist_len[filtered_users]

            lowest_degree = torch.argsort(hist_len)[:int(self.users_low_degree_ratio * hist_len.shape[0])]
            filtered_users = filtered_users[lowest_degree]

        # sparse users are connected to low-degree items
        if self.sparse_users_constraint and self.random_perturbation:
            raise NotImplementedError(
                'The policies `sparse_users_constraint` and `random_perturbation` cannot be both True'
            )
        elif self.sparse_users_constraint:
            if filtered_users is None:
                filtered_users = batched_data

            sparsity_df = eval_utils.extract_graph_metrics_per_node(
                self.dataset, remove_first_row_col=True, metrics=["Sparsity"]
            )
            sparsity = torch.from_numpy(
                sparsity_df.set_index('Node').loc[filtered_users.numpy(), 'Sparsity'].to_numpy()
            )

            most_sparse = torch.argsort(sparsity)[-int(self.sparse_users_constraint_ratio * sparsity.shape[0]):]
            filtered_users = filtered_users[most_sparse]

        if self.items_preference_constraint:
            ihist_m, _, ihist_len = self.dataset.history_user_matrix()

            sens_map = self.dataset.user_feat[self.sensitive_attribute]
            n_adv = (sens_map == self.adv_group).sum() / (sens_map.shape[0] - 1)

            sens_ihist_m = sens_map[ihist_m]
            adv_ratio = torch.nan_to_num((sens_ihist_m == self.adv_group).sum(dim=1) / ihist_len, nan=0)
            # values higher than 1 means the advantaged group prefer those items w.r.t. to their representation
            adv_ratio = adv_ratio / n_adv

            filtered_items = torch.argsort(adv_ratio)[-int(self.items_preference_constraint_ratio * adv_ratio.shape[0]):]

        if self.niche_items_constraint:
            if filtered_items is None:
                filtered_items = torch.arange(1, self.dataset.item_num)
            _, _, ihist_len = self.dataset.history_user_matrix()
            ihist_len = ihist_len[filtered_items]

            filtered_items = filtered_items[
                torch.argsort(ihist_len)[:int(self.niche_items_constraint_ratio * ihist_len.shape[0])]
            ]

        return batched_data, filtered_users, filtered_items, (test_model_topk, test_scores_args, rec_model_topk, rec_scores_args)

    def _check_previous_loss_value(self):
        previous_loss_value = None
        if self.previous_loss_value:
            previous_loss_value = dict.fromkeys(self.dataset.user_feat[self.sensitive_attribute].unique().numpy())
            for gr in previous_loss_value:
                previous_loss_value[gr] = np.array([])

        return previous_loss_value

    def _check_early_stopping(self, check_value, epoch, *update_best_example_args):
        if self.earlys.check(check_value):
            self.logger.info(self.earlys)
            best_epoch = epoch + 1 - self.earlys.patience
            self.logger.info(f"Early Stopping: best epoch {best_epoch}")

            # stub example added to find again the best epoch when explanations are loaded
            self.update_best_cf_example(*update_best_example_args, force_update=True)

            return True
        return False

    def _check_item_feat(self):
        if self.dataset.item_feat is None or self.item_discriminative_attribute not in self.dataset.item_feat:
            utils.update_item_feat_discriminative_attribute(
                self.dataset, self.item_discriminative_attribute, self.item_discriminative_ratio, self.item_discriminative_map
            )

    def _pref_data_sens_and_metric(self, pref_users, model_topk, eval_data=None):
        pref_data = pd.DataFrame(
            zip(pref_users, model_topk, self.dataset.user_feat[self.sensitive_attribute][pref_users].numpy()),
            columns=['user_id', 'topk_pred', 'Demo Group']
        )
        pref_data[self.eval_metric] = self.compute_eval_metric(eval_data or self.rec_data.dataset, pref_data, 'topk_pred')[:, -1]

        return pref_data

    def compute_eval_metric(self, dataset, pref_data, col):
        return eval_utils.compute_metric(self.evaluator, dataset, pref_data, col, self.eval_metric)

    def compute_eval_result(self, pref_users: np.ndarray, model_topk: np.ndarray, eval_data=None):
        return self._pref_data_sens_and_metric(pref_users, model_topk, eval_data=eval_data)[self.eval_metric].to_numpy()

    def compute_f_m_result(self, batched_data: np.ndarray, model_topk, eval_data=None):
        pref_data = self._pref_data_sens_and_metric(batched_data, model_topk, eval_data=eval_data)

        f_result = pref_data.loc[pref_data['Demo Group'] == self.f_idx, self.eval_metric].mean()
        m_result = pref_data.loc[pref_data['Demo Group'] == self.m_idx, self.eval_metric].mean()

        return f_result, m_result

    def compute_exp_metric(self, pref_users, model_topk, dataset, iterations=100):
        pref_data = self._pref_data_sens_and_metric(pref_users, model_topk, eval_data=dataset)

        return exp_metrics.get_exp_metric_value(
            self.exp_metric,
            pref_data,
            self.eval_metric,
            dataset.dataset_name,
            self.sensitive_attribute,
            self.user_batch_exp,
            dataset=self.dataset,
            discriminative_attribute=self.item_discriminative_attribute,
            groups_distrib=self.item_discriminative_groups_distrib,
            iterations=iterations,
            coverage_min_relevant_items=self.coverage_min_relevant_items
        )

    def determine_adv_group(self, batched_data: np.ndarray, rec_model_topk):
        f_result, m_result = self.compute_f_m_result(batched_data, rec_model_topk)

        check_func = "__ge__" if self.config['perturb_adv_group'] else "__lt__"

        self.adv_group = self.m_idx if getattr(m_result, check_func)(f_result) else self.f_idx
        self.disadv_group = self.f_idx if self.adv_group == self.m_idx else self.m_idx
        self.results = dict(zip([self.m_idx, self.f_idx], [m_result, f_result]))

    def increase_dataset_unfairness(self, batched_data, test_data, rec_model_topk, topk=10):
        pref_users = batched_data.numpy()
        pref_data = pd.DataFrame(zip(pref_users, rec_model_topk.tolist()), columns=['user_id', 'topk_pred'])

        orig_res = self.compute_eval_metric(self.rec_data.dataset, pref_data, 'topk_pred')

        m_users_mask = self.dataset.user_feat[self.sensitive_attribute][pref_users] == self.m_idx
        f_users_mask = self.dataset.user_feat[self.sensitive_attribute][pref_users] == self.f_idx

        m_size, f_size = m_users_mask.nonzero().shape[0], f_users_mask.nonzero().shape[0]
        if m_size >= f_size:
            gr_to_reduce, gr_fixed = self.m_idx, self.f_idx
            steps = np.linspace(f_size, m_size, 10, dtype=int)[::-1]
        else:
            gr_to_reduce, gr_fixed = self.f_idx, self.m_idx
            steps = np.linspace(m_size, f_size, 10, dtype=int)[::-1]

        df_res = pd.DataFrame(zip(
            pref_users,
            orig_res[:, - 1],
            self.dataset.user_feat[self.sensitive_attribute][pref_users].numpy()
        ), columns=['user_id', 'result', self.sensitive_attribute])

        if self.dataset.dataset_name == "lastfm-1k":
            ascending = True
            def check_func(gr_red_res, gr_fix_res): return gr_red_res <= gr_fix_res / 2
        else:
            ascending = False
            def check_func(gr_red_res, gr_fix_res): return gr_fix_res <= gr_red_res / 2

        for step in steps:
            step_df = df_res.groupby(self.sensitive_attribute).apply(
                lambda x: x.sort_values('result', ascending=ascending)[:step]
            ).reset_index(drop=True)
            mean_metric = step_df.groupby(self.sensitive_attribute).mean()

            batched_data = torch.tensor(step_df['user_id'].to_numpy())

            if check_func(mean_metric.loc[gr_to_reduce, 'result'], mean_metric.loc[gr_fixed, 'result']):
                print(mean_metric)
                break

        # recompute recommendations of original model
        test_scores_args, test_model_topk = self._get_model_score_data(batched_data, test_data, topk)
        rec_scores_args, rec_model_topk = self._get_model_score_data(batched_data, self.rec_data, topk)

        return batched_data, test_model_topk, test_scores_args, rec_model_topk, rec_scores_args

    def run_epoch(self, epoch, batched_data, exp_losses, topk=10):
        iter_data = self.prepare_iter_batched_data(batched_data)
        iter_data = [iter_data[i] for i in np.random.permutation(len(iter_data))]

        previous_loss_value = self._check_previous_loss_value()
        if self._exp_loss.loss_type() == 'Consumer':
            self._exp_loss.update_previous_loss_value(previous_loss_value)

        if not self.mini_batch_descent:
            self.cf_optimizer.zero_grad()

        epoch_exp_loss = []
        new_example, loss_total = None, None
        for batch_idx, batch_user in enumerate(iter_data):
            if self.previous_batch_LR_scaling:
                self.lr_scaler.update(batch_idx)

            batch_scores_args = self._get_scores_args(batch_user, self.rec_data)

            torch.cuda.empty_cache()
            new_example, loss_total, exp_loss = self.train(
                epoch, batch_scores_args, batch_user, topk=topk, previous_loss_value=previous_loss_value
            )
            epoch_exp_loss.append(exp_loss)

            if batch_idx != len(iter_data) - 1:
                torch.nn.utils.clip_grad_norm_(self.cf_model.parameters(), 2.0)
                if self.mini_batch_descent:
                    self.cf_optimizer.step()

        if self.previous_batch_LR_scaling:
            self.lr_scaler.restore()

        epoch_exp_loss = np.mean(epoch_exp_loss)
        exp_losses.append(epoch_exp_loss)

        return new_example, loss_total, epoch_exp_loss

    def explain(self, batched_data, full_dataset, train_data, valid_data, test_data, epochs, topk=10):
        """
        The method from which starts the perturbation of the graph by optimization of `pred_loss` or `exp_loss`
        :param batched_data:
        :param test_data:
        :param epochs:
        :param topk:
        :return:
        """
        best_loss = np.inf

        self._check_loss_trend_epoch_images()

        test_scores_args, test_model_topk = self._get_model_score_data(batched_data, test_data, topk)

        # recommendations generated by the model are considered the ground truth
        if self._pred_as_rec:
            rec_scores_args, rec_model_topk = self._prepare_test_history_matrix(test_data, topk=topk)
        else:
            rec_scores_args, rec_model_topk = self._get_model_score_data(batched_data, self.rec_data, topk)

        batched_data, filtered_users, filtered_items, inc_disp_model_data = self._check_policies(batched_data, rec_model_topk)
        if self.increase_disparity:
            test_model_topk, test_scores_args, rec_model_topk, rec_scores_args = inc_disp_model_data

        self._exp_loss = self._exp_loss(
            *self._exp_loss_args[self.exp_metric.lower()],
            topk=topk,
            loss=self._metric_loss,
            adv_group_data=(self.only_adv_group, self.disadv_group, self.results[self.disadv_group]),
            deactivate_gradient=self.gradient_deactivation_constraint,
            only_relevant=self.coverage_loss_only_relevant,
            groups_distrib=self.item_discriminative_groups_distrib
        )

        if self._exp_loss.loss_type() == 'Provider':
            self._check_item_feat()

        # logs of explanation consider validation as seen when model recommendations are used as ground truth
        if self._pred_as_rec:
            self.rec_data = test_data

        detached_batched_data = batched_data.detach().numpy()
        self.initialize_cf_model(filtered_users=filtered_users, filtered_items=filtered_items)
        self.initialize_optimizer()

        if self.ckpt_loading_path is not None and os.path.exists(self.ckpt_loading_path):
            exp_losses, starting_epoch, best_cf_example = self._resume_checkpoint()
            last_earlys_check_value = self.earlys.history.pop()
            if self.earlys.check(last_earlys_check_value):
                raise AttributeError("A checkpoint of a completed run cannot be resumed")
        else:
            starting_epoch = 0
            exp_losses = []
            best_cf_example = []

        # orig_rec_dp = eval_utils.compute_DP(
        #     self.results[self.adv_group], self.results[self.disadv_group]
        # )
        # orig_test_dp = eval_utils.compute_DP(
        #     *self.compute_f_m_result(detached_batched_data, test_model_topk, eval_data=test_data.dataset)
        # )

        orig_rec_exp_metric_val = self.compute_exp_metric(detached_batched_data, rec_model_topk, self.dataset)
        orig_rec_exp_metric_test = self.compute_exp_metric(detached_batched_data, test_model_topk, test_data.dataset)

        self.logger.info("*********** Rec Data ***********")
        if self._exp_loss.loss_type() == 'Consumer':
            self.logger.info(self.results)
            self.logger.info(f"M idx: {self.m_idx}")
        self.logger.info(f"Original Rec Explanation Metric Value: {orig_rec_exp_metric_val}")
        self.logger.info("*********** Test Data ***********")
        self.logger.info(f"Original Test Explanation Metric Value: {orig_rec_exp_metric_test}")

        iter_epochs = tqdm(
            range(starting_epoch, epochs),
            total=epochs,
            ncols=100,
            initial=starting_epoch,
            desc=set_color(f"Epochs   ", 'blue'),
        )

        if self.previous_batch_LR_scaling:
            iter_data = self.prepare_iter_batched_data(batched_data)
            self.lr_scaler = exp_utils.LRScaler(self.cf_optimizer, len(iter_data))

        for epoch in iter_epochs:
            new_example, loss_total, epoch_exp_loss = self.run_epoch(epoch, batched_data, exp_losses, topk=topk)

            if self.verbose:
                Explainer._verbose_plot(exp_losses, epoch)

            if new_example is not None:
                new_example, epoch_exp_metric = self.update_new_example(
                    new_example,
                    detached_batched_data,
                    full_dataset,
                    train_data,
                    valid_data,
                    test_data,
                    test_model_topk,
                    rec_model_topk,
                    epoch_exp_loss
                )

                if new_example is False:
                    self.logger.warning('Interrupting because one of the users interacted with all the items')
                    break

                earlys_check_value = {
                    'exp_loss': epoch_exp_loss,
                    'exp_metric': epoch_exp_metric
                }[self.earlys_check_value]
                if self._pred_as_rec and earlys_check_value == epoch_exp_metric:
                    raise ValueError(f"`exp_rec_data` = `rec` stores test data to log explanation metric. "
                                     f"Cannot be used as value for early stopping check")

                update_best_example_args = [best_cf_example, new_example, loss_total, best_loss]
                earlys_check = self._check_early_stopping(earlys_check_value, epoch, *update_best_example_args)
                print("*" * 7 + " Early Stopping History " + "*" * 7)
                print(self.earlys.history)
                print("*" * 7 + "************************" + "*" * 7)
                if self.ckpt_loading_path is not None:
                    if not earlys_check:
                        # epoch + 1 because the current one is already finished
                        self._save_checkpoint(epoch + 1, exp_losses, best_cf_example)

                if earlys_check:
                    break
                elif epoch == (epochs - 1):
                    # stub explanation that means the code stopped because of the number of epochs limit, not because of the early stopping
                    self.update_best_cf_example(best_cf_example, utils._EXPS_END_EPOCHS_STUB, loss_total, best_loss, model_topk=rec_model_topk)
                    break

                best_loss = self.update_best_cf_example(*update_best_example_args, model_topk=rec_model_topk)

            # the optimizer step of the last epoch_step is done here to prevent computations
            # done to update new example to be related to the new state of the model
            torch.nn.utils.clip_grad_norm_(self.cf_model.parameters(), 2.0)
            if self.mini_batch_descent:
                self.cf_optimizer.step()

            self.logger.info("{} CF examples".format(len(best_cf_example)))

        return best_cf_example, detached_batched_data, (rec_model_topk, test_model_topk)

    def train(self, epoch, scores_args, users_ids, topk=10, previous_loss_value=None):
        """
        Training procedure of explanation
        :param epoch:
        :param topk:
        :return:
        """
        t = time.time()

        # `_get_no_grad_pred_model_score_data` call torch `eval` inside
        cf_topk_pred_idx, _ = self._get_no_grad_pred_model_score_data(scores_args)

        if self.mini_batch_descent:
            self.cf_optimizer.zero_grad()
        self.cf_model.train()

        # compute differentiable permutation of adj matrix
        # cf_scores uses differentiable P_hat ---> adjacency matrix not binary, but needed for training
        cf_scores = self.get_scores(self.cf_model, *scores_args, pred=False)

        # remove neginf from output
        cf_scores = torch.nan_to_num(cf_scores, neginf=(torch.min(cf_scores[~torch.isinf(cf_scores)]) - 1).item())
        cf_scores_topk, cf_topk_idx = self.get_top_k(cf_scores, **self.topk_args)

        user_feat = self.get_batch_user_feat(users_ids)
        target = self.get_target(cf_scores, user_feat)

        if self._exp_loss.is_data_feat_needed():
            data_feat = self.get_loss_data_feat(users_ids, user_feat)
            self._exp_loss.update_data_feat(data_feat)

        loss_total, orig_loss_graph_dist, loss_graph_dist, exp_loss, adj_sub_cf_adj = self.cf_model.loss(
            cf_scores,
            self._exp_loss,
            target
        )

        torch.cuda.empty_cache()
        loss_total.backward()

        # for name, param in self.cf_model.named_parameters():
        #     if name == "P_symm":
        #         print(param.grad)
        # import pdb; pdb.set_trace()

        exp_loss = exp_loss.mean().item() if exp_loss is not None else torch.nan
        self.log_epoch(
            t, epoch, loss_total, exp_loss, loss_graph_dist, orig_loss_graph_dist,
            **dict(cf_topk_idx=cf_topk_idx, cf_topk_pred_idx=cf_topk_pred_idx)
        )

        cf_stats = None
        if orig_loss_graph_dist.item() > 0:
            cf_stats = self.get_batch_cf_stats(
                adj_sub_cf_adj, loss_total, loss_graph_dist, exp_loss, epoch
            )

        return cf_stats, loss_total.item(), exp_loss

    def get_batch_user_feat(self, users_ids):
        user_feat = self.dataset.user_feat
        user_id_mask = users_ids.unsqueeze(-1) if users_ids.dim() == 0 else users_ids
        return {k: feat[user_id_mask] for k, feat in user_feat.interaction.items()}

    def get_target(self, cf_scores, user_feat):
        target = torch.zeros_like(cf_scores, dtype=torch.float, device=cf_scores.device)

        if not self._pred_as_rec:
            hist_matrix, _, _ = self.rec_data.dataset.history_item_matrix()
            rec_data_interactions = hist_matrix[user_feat[self.dataset.uid_field]]
        else:
            rec_data_interactions = self._test_history_matrix[user_feat[self.dataset.uid_field]]
        target[torch.arange(target.shape[0])[:, None], rec_data_interactions] = 1
        target[:, 0] = 0  # item 0 is a padding item

        return target

    def get_loss_data_feat(self, users_ids, user_feat):
        if self._exp_loss.loss_type() == 'Consumer':
            return user_feat
        elif self._exp_loss.loss_type() == 'Provider':
            return self.dataset.item_feat[1:]
        else:
            raise ValueError('A `FairLoss` subclass should contain "Consumer" or "Provider" in its name')

    def get_batch_cf_stats(self, adj_sub_cf_adj, loss_total, loss_graph_dist, exp_loss, epoch):
        # Compute distance between original and perturbed list. Explanation maintained only if dist > 0
        # cf_dist = [self.dist(_pred, _topk_idx) for _pred, _topk_idx in zip(cf_topk_pred_idx, self.model_topk_idx)]
        cf_dist = None

        adj_pert_edges = adj_sub_cf_adj.detach().cpu()
        pert_edges = adj_pert_edges.indices()[:, adj_pert_edges.values().nonzero().squeeze()]

        # remove duplicated edges
        pert_edges = pert_edges[:, (pert_edges[0, :] < self.dataset.user_num) & (pert_edges[0, :] > 0)].numpy()

        cf_stats = [loss_total.item(), loss_graph_dist.item(), exp_loss, None, pert_edges, epoch + 1]

        if self.neighborhood_perturbation:
            self.cf_model.update_neighborhood(torch.Tensor(pert_edges))

        return cf_stats

    @staticmethod
    def get_scores(_model, batched_data, tot_item_num, test_batch_size, item_tensor, pred=False):
        return exp_utils.get_scores(_model, batched_data, tot_item_num, test_batch_size, item_tensor, pred=pred)

    @staticmethod
    def get_top_k(scores_tensor, topk=10):
        return utils.get_top_k(scores_tensor, topk=topk)



import os
import copy
import stat
import shutil
import pickle
from logging import getLogger

import yaml
import wandb
import torch
import scipy
import numba
import numpy as np
import pandas as pd
from recbole.config import Config
from sklearn.decomposition import PCA
from recbole.data.interaction import Interaction
from recbole.utils import init_logger, get_model, FeatureSource, FeatureType
from recbole.data import data_preparation, create_samplers, get_dataloader

from gnnuers.data import Dataset

try:
    from torch_geometric.utils import k_hop_subgraph, subgraph
except ModuleNotFoundError:
    pass


_EXPS_END_EPOCHS_STUB = ['STUB', 'STUB']


_EXPS_COLUMNS = [
    # "user_id",
    # "rec_topk",
    # "test_topk",
    # "rec_cf_topk",
    # "test_cf_topk",
    # "rec_cf_dist",
    # "test_cf_dist",
    "loss_total",
    "loss_graph_dist",
    "exp_loss",
    "exp_metric",
    "del_edges",
    "epoch",
    # "first_fair_loss"
]


_OLD_RS23_TIST_EXPS_COLUMNS = [
    "user_id",
    # "rec_topk",
    # "test_topk",
    "rec_cf_topk",
    "test_cf_topk",
    "rec_cf_dist",
    "test_cf_dist",
    "loss_total",
    "loss_graph_dist",
    "fair_loss",
    "fair_metric",
    "del_edges",
    "epoch",
    "first_fair_loss"
]


_OLD_EXPS_COLUMNS = [
    "user_id",
    "rec_topk",
    "test_topk",
    "rec_cf_topk",
    "test_cf_topk",
    "rec_cf_dist",
    "test_cf_dist",
    "loss_total",
    "loss_graph_dist",
    "fair_loss",
    "del_edges",
    "epoch",
    "first_fair_loss"
]


def exp_col_index(col):
    try:
        idx = _EXPS_COLUMNS.index(col)
    except ValueError:
        idx = col
    return idx


def old_exp_col_index(col):
    try:
        idx = _OLD_EXPS_COLUMNS.index(col)
    except ValueError:
        idx = col
    return idx


def old_rs23_tist_exp_col_index(col):
    try:
        idx = _OLD_RS23_TIST_EXPS_COLUMNS.index(col)
    except ValueError:
        idx = col
    return idx


def wandb_init(config, policies=None, **kwargs):
    config = config.final_config_dict if not isinstance(config, dict) else config

    tags = None
    policies = config.get("explainer_policies", policies)
    if policies is not None:
        tags = [k for k in policies if policies[k]]
    config['wandb_tags'] = tags

    return wandb.init(
        **kwargs,
        tags=tags,
        config=config
    )


def load_data_and_model(model_file, explainer_config=None, cmd_config_args=None, perturbed_dataset=None):
    r"""Load filtered dataset, split dataloaders and saved model.
    Args:
        model_file (str): The path of saved model file.
    Returns:
        tuple:
            - config (Config): An instance object of Config, which record parameter information in :attr:`model_file`.
            - model (AbstractRecommender): The model load from :attr:`model_file`.
            - dataset (Dataset): The filtered dataset.
            - train_data (AbstractDataLoader): The dataloader for training.
            - valid_data (AbstractDataLoader): The dataloader for validation.
            - test_data (AbstractDataLoader): The dataloader for testing.
    """
    checkpoint = torch.load(model_file)
    config = checkpoint['config']

    if explainer_config is not None:
        if isinstance(explainer_config, str):
            with open(explainer_config, 'r', encoding='utf-8') as f:
                explain_config_dict = yaml.load(f.read(), Loader=config.yaml_loader)
        elif isinstance(explainer_config, dict):
            explain_config_dict = explainer_config
        else:
            raise ValueError(f'explainer_config cannot be `{type(explainer_config)}`. Only `str` and `dict` are supported')

        config.final_config_dict.update(explain_config_dict)

    if cmd_config_args is not None:
        for arg, val in cmd_config_args.items():
            conf = config
            if '.' in arg:
                subargs = arg.split('.')
                for subarg in subargs[:-1]:
                    conf = conf[subarg]
                arg = subargs[-1]

            if conf[arg] is None:
                try:
                    new_val = float(val)
                    new_val = int(new_val) if new_val.is_integer() else new_val
                except ValueError:
                    new_val = int(val) if val.isdigit() else val
                conf[arg] = new_val
            else:
                try:
                    arg_type = type(conf[arg])
                    if arg_type == bool:
                        new_val = val.title() == 'True'
                    else:
                        new_val = arg_type(val)  # cast to same type in config
                    conf[arg] = new_val
                except (ValueError, TypeError):
                    new_val = None

    config['data_path'] = config['data_path'].replace('\\', os.sep)
    # config['device'] = torch.device('cuda')

    logger = getLogger()
    logger.info(config)

    if perturbed_dataset is not None:
        dataset = perturbed_dataset
    else:
        dataset = Dataset(config)
        default_file = os.path.join(config['checkpoint_dir'], f'{config["dataset"]}-{dataset.__class__.__name__}.pth')
        file = config['dataset_save_path'] or default_file
        if os.path.exists(file):
            with open(file, 'rb') as f:
                dataset = pickle.load(f)
            dataset_args_unchanged = True
            for arg in dataset_arguments + ['seed', 'repeatable']:
                if config[arg] != dataset.config[arg]:
                    dataset_args_unchanged = False
                    break
            if dataset_args_unchanged:
                logger = getLogger()
                logger.info(set_color('Load filtered dataset from', 'pink') + f': [{file}]')

        if config['save_dataset']:
            dataset.save()

    logger.info(dataset)

    train_data, valid_data, test_data = data_preparation(config, dataset)

    model = get_model(config['model'])(config, train_data.dataset).to(config['device'])
    model.load_state_dict(checkpoint['state_dict'])
    model.load_other_parameter(checkpoint.get('other_parameter'))

    logger.info(model)

    return config, model, dataset, train_data, valid_data, test_data


def update_base_explainer(base_explainer_config_file, explainer_config_file=None):
    yaml_loader = Config('GCMC', 'ml-1m').yaml_loader  # random attributes just to retrieve the yaml_loader

    with open(base_explainer_config_file, 'r', encoding='utf-8') as f:
        config_dict = yaml.load(f.read(), Loader=yaml_loader)

    if explainer_config_file is not None:
        if os.path.splitext(explainer_config_file)[-1] == '.yaml':
            with open(explainer_config_file, 'r', encoding='utf-8') as f:
                exp_config_dict = yaml.load(f.read(), Loader=yaml_loader)
        else:
            with open(explainer_config_file, 'rb') as f:
                exp_config_dict = pickle.load(f).final_config_dict

        config_dict.update(exp_config_dict)

    return config_dict


def load_dp_exps_file(base_exps_file):
    cf_data_file = os.path.join(base_exps_file, 'cf_data.pkl')
    model_preds_file = os.path.join(base_exps_file, 'model_rec_test_preds.pkl')

    with open(cf_data_file, 'rb') as file:
        exps = [pickle.load(file)]
    with open(model_preds_file, 'rb') as file:
        model_preds = pickle.load(file)

    return exps, *model_preds


def load_old_dp_exps_file(base_exps_file):
    cf_data_file = os.path.join(base_exps_file, 'all_users.pkl')

    with open(cf_data_file, 'rb') as file:
        exps = [pickle.load(file)]

    return exps


def get_dataset_with_perturbed_edges(pert_edges: np.ndarray, data_set):
    user_num = data_set.user_num
    uid_field, iid_field = data_set.uid_field, data_set.iid_field
    pert_edges = pert_edges.copy()

    pert_edges = torch.tensor(pert_edges)
    pert_edges[1] -= user_num  # remap items in range [0, item_num)

    orig_inter_feat = data_set.inter_feat
    pert_inter_feat = {}
    for i, col in enumerate([uid_field, iid_field]):
        pert_inter_feat[col] = torch.cat((orig_inter_feat[col], pert_edges[i]))

    unique, counts = torch.stack(
        (pert_inter_feat[uid_field], pert_inter_feat[iid_field]),
    ).unique(dim=1, return_counts=True)
    pert_inter_feat[uid_field], pert_inter_feat[iid_field] = unique[:, counts == 1]

    return data_set.copy(Interaction(pert_inter_feat))


def get_dataloader_with_perturbed_edges(pert_edges: np.ndarray, config, dataset, train_data, valid_data, test_data):
    pert_edges = pert_edges.copy()

    train_dataset = get_dataset_with_perturbed_edges(pert_edges, train_data.dataset)
    valid_dataset = get_dataset_with_perturbed_edges(pert_edges, valid_data.dataset)
    test_dataset = get_dataset_with_perturbed_edges(pert_edges, test_data.dataset)

    built_datasets = [train_dataset, valid_dataset, test_dataset]
    train_sampler, valid_sampler, test_sampler = create_samplers(config, dataset, built_datasets)

    train_data = get_dataloader(config, 'train')(config, train_dataset, train_sampler, shuffle=False)
    valid_data = get_dataloader(config, 'evaluation')(config, valid_dataset, valid_sampler, shuffle=False)
    test_data = get_dataloader(config, 'evaluation')(config, test_dataset, test_sampler, shuffle=False)

    return train_data, valid_data, test_data


def get_best_exp_early_stopping(exps, config_dict):
    if exps[-1] == _EXPS_END_EPOCHS_STUB:
        return exps[-2]

    best_epoch = get_best_epoch_early_stopping(exps, config_dict)
    epoch_idx = exp_col_index('epoch')

    return [e for e in sorted(exps, key=lambda x: abs(x[epoch_idx] - best_epoch)) if e[epoch_idx] <= best_epoch][0]


def get_exp_by_epoch(exps, query_epoch):
    epoch_idx = exp_col_index('epoch')

    queried_exp = [e for e in exps if e[epoch_idx] == query_epoch]

    return queried_exp[0] if len(queried_exp) > 0 else None


def old_get_best_epoch_early_stopping(exps, config_dict):
    try:
        patience = config_dict['early_stopping']['patience']
    except TypeError:
        patience = config_dict['earlys_patience']

    old_index_func = old_rs23_tist_exp_col_index if len(exps[0]) == 12 else old_exp_col_index

    return max([e[old_index_func('epoch')] for e in exps]) - patience


def get_best_epoch_early_stopping(exps, config_dict):
    try:
        patience = config_dict['early_stopping']['patience']
    except TypeError:
        patience = config_dict['earlys_patience']

    if exps[-1] == _EXPS_END_EPOCHS_STUB:
        return exps[-2][exp_col_index('epoch')]

    max_epoch = max([e[exp_col_index('epoch')] for e in exps])
    # the training process stopped because of other condition
    if max_epoch <= patience:
        return max_epoch

    return max_epoch - patience


def prepare_batched_data(input_data, data, item_data=None):
    """
    Prepare the batched data according to the "recbole" pipeline
    :param batched_data:
    :param data:
    :param item_data:
    :return:
    """
    data_df = Interaction({k: v[input_data] for k, v in data.dataset.user_feat.interaction.items()})

    if item_data is not None:
        data_df.update(Interaction({data.dataset.iid_field: item_data}))

    if hasattr(data, "uid2history_item"):
        history_item = data.uid2history_item[data_df[data.dataset.uid_field]]
        history_item = [x if x is not None else torch.Tensor([]) for x in history_item]

        if len(input_data) > 1:
            history_u = torch.cat([torch.full_like(hist_iid, i, dtype=int) for i, hist_iid in enumerate(history_item)])
            history_i = torch.cat(list(history_item)).long()
        else:
            history_u = torch.full_like(history_item, 0)
            history_i = history_item

        history_index = (history_u, history_i)
    else:
        history_index = None

    return data_df, history_index, None, None


def _spilt_predict(model, interaction, batch_size, test_batch_size):
    spilt_interaction = dict()
    for key, tensor in interaction.interaction.items():
        spilt_interaction[key] = tensor.split(test_batch_size, dim=0)
    num_block = (batch_size + test_batch_size - 1) // test_batch_size
    result_list = []
    for i in range(num_block):
        current_interaction = dict()
        for key, spilt_tensor in spilt_interaction.items():
            current_interaction[key] = spilt_tensor[i]
        result = model.predict(Interaction(current_interaction).to(model.device))
        if len(result.shape) == 0:
            result = result.unsqueeze(0)
        result_list.append(result)
    return torch.cat(result_list, dim=0)


def get_scores(model, batched_data, tot_item_num, test_batch_size, item_tensor, **kwargs):
    interaction, history_index, _, _ = batched_data
    inter_data = interaction.to(model.device)
    try:
        scores = model.full_sort_predict(inter_data, **kwargs)

    except NotImplementedError:
        inter_len = len(interaction)
        new_inter = interaction.to(model.device, **scores_kws).repeat_interleave(tot_item_num)
        batch_size = len(new_inter)
        new_inter.update(item_tensor.repeat(inter_len))
        if batch_size <= test_batch_size:
            scores = model.predict(new_inter)
        else:
            scores = _spilt_predict(model, new_inter, batch_size, test_batch_size)

    scores = scores.view(-1, tot_item_num)
    scores[:, 0] = -np.inf
    if model.ITEM_ID in interaction:
        scores = scores[:, inter_data[model.ITEM_ID]]
    if history_index is not None:
        scores[history_index] = -np.inf

    return scores


def get_top_k(scores_tensor, topk=10):
    scores_top_k, topk_idx = torch.topk(scores_tensor, topk, dim=-1)  # n_users x k

    return scores_top_k, topk_idx


def update_item_feat_discriminative_attribute(dataset, item_discriminative_attribute, item_discriminative_ratio, item_discriminative_map):
    if isinstance(dataset.inter_feat, pd.DataFrame):
        # + 1 because the inter_feat does not contain the padding user and item, so the indices need to be re-indexed by 1
        pop = torch.tensor(dataset.inter_feat[dataset.iid_field].value_counts(sort=False).argsort()[::-1].values + 1)
    else:
        pop = torch.argsort(dataset.history_user_matrix()[2], descending=True)
    sh_size = round(item_discriminative_ratio * pop.shape[0])

    if dataset.item_feat is None:
        dataset.item_feat = dataset.get_item_feature()
    # field2seqlen[dest_field]

    exposure_group = torch.zeros_like(dataset.item_feat[dataset.iid_field])
    exposure_group[pop[:sh_size]] = 1
    exposure_group[pop[sh_size:]] = 2
    exposure_group[0] = 0

    dataset.item_feat[item_discriminative_attribute] = exposure_group
    dataset.field2type[item_discriminative_attribute] = FeatureType('token')
    dataset.field2source[item_discriminative_attribute] = FeatureSource('item')
    dataset.field2seqlen[item_discriminative_attribute] = 1
    dataset.field2id_token[item_discriminative_attribute] = np.asarray(item_discriminative_map)


def get_decomposed_adj_matrix(del_edges, train_dataset, method='PCA'):
    pert_train_dataset = get_dataset_with_perturbed_edges(del_edges, train_dataset)
    train_adj = train_dataset.inter_matrix(form='csr').astype(np.float32)[1:, 1:].todense()
    pert_train_adj = pert_train_dataset.inter_matrix(form='csr').astype(np.float32)[1:, 1:].todense()

    if method.upper() == "PCA":
        decomposer = PCA(n_components=2)
    else:
        raise NotImplementedError("Only PCA is a supported decomposer")

    train_pca = decomposer.fit_transform(train_adj)
    pert_train_pca = decomposer.fit_transform(pert_train_adj)

    return train_pca, pert_train_pca


def get_degree_matrix(adj):
    return torch.diag(adj.sum(dim=1))


def get_neighbourhood(node_idx,
                      edge_index,
                      n_hops,
                      neighbors_hops=False,
                      only_last_level=False,
                      not_user_sub_matrix=False,
                      max_num_nodes=None):
    def hop_difference(_node_idx, _edge_index, edge_s, _n_hops):
        _edge_subset = k_hop_subgraph(_node_idx, _n_hops, _edge_index)
        _edge_subset = subgraph(_edge_subset[0], _edge_index)  # Get subset of edges

        # takes the non-intersection between last level subgraph and lower hop subgraph
        unique, counts = torch.cat((edge_s[0], _edge_subset[0]), dim=1).unique(dim=1, return_counts=True)
        edge_s = [unique[:, counts == 1]]
        return edge_s

    if neighbors_hops:
        n_hops = n_hops * 2

    edge_subset = k_hop_subgraph(node_idx, n_hops, edge_index)  # Get all nodes involved
    edge_subset = subgraph(edge_subset[0], edge_index)  # Get subset of edges
    if n_hops > 1:
        if only_last_level:
            edge_subset = hop_difference(node_idx, edge_index, edge_subset, n_hops - 1)
        if not_user_sub_matrix:
            edge_subset = hop_difference(node_idx, edge_index, edge_subset, 1)

    # sub_adj = to_dense_adj(edge_subset[0], max_num_nodes=max_num_nodes).squeeze().to_sparse()

    # return sub_adj, edge_subset
    return edge_subset


def is_symmetrically_sorted(idxs: torch.Tensor):
    assert idxs.shape[0] == 2
    symm_offset = idxs.shape[1] // 2
    return (idxs[:, :symm_offset] == idxs[[1, 0], symm_offset:]).all()


def get_sorter_indices(base_idxs, to_sort_idxs):
    unique, inverse = torch.cat((base_idxs, to_sort_idxs), dim=1).unique(dim=1, return_inverse=True)
    inv_base, inv_to_sort = torch.split(inverse, to_sort_idxs.shape[1])
    sorter = torch.arange(to_sort_idxs.shape[1], device=inv_to_sort.device)[torch.argsort(inv_to_sort)]

    return sorter, inv_base


def create_symm_matrix_from_vec(vector, n_rows, idx=None, base_symm='zeros'):
    if isinstance(base_symm, str):
        symm_matrix = getattr(torch, base_symm)(n_rows, n_rows).to(vector.device)
    else:
        symm_matrix = copy.deepcopy(base_symm)
    old_idx = idx
    if old_idx is None:
        idx = torch.tril_indices(n_rows, n_rows, -1)
    symm_matrix[idx[0], idx[1]] = vector
    symm_matrix[idx[1], idx[0]] = vector

    return symm_matrix


def create_sparse_symm_matrix_from_vec(vector, idx, base_symm, edge_deletions=False, mask_filter=None):
    symm_matrix = copy.deepcopy(base_symm)
    if not symm_matrix.is_coalesced():
        symm_matrix = symm_matrix.coalesce()
    symm_matrix_idxs, symm_matrix_vals = symm_matrix.indices(), symm_matrix.values()

    if not edge_deletions:  # if pass is edge additions
        vector_zeros_mask = vector != 0

        idx = idx[:, vector_zeros_mask]
        vector = vector[vector_zeros_mask]
        del vector_zeros_mask
        torch.cuda.empty_cache()

        symm_matrix_idxs = symm_matrix_idxs.to('cuda:0')
        idx = idx.to('cuda:0')
        idx = torch.cat((symm_matrix_idxs, idx, idx[[1, 0]]), dim=1)
        torch.cuda.empty_cache()
        vector = torch.cat((symm_matrix_vals, vector, vector))
        torch.cuda.empty_cache()
        idx = idx.to(vector.device)
        torch.cuda.empty_cache()
    else:
        vector = torch.cat((vector, vector))
        if mask_filter is not None:
            sorter, idx_inverse = get_sorter_indices(idx, symm_matrix_idxs)
            symm_matrix_vals = symm_matrix_vals[sorter][idx_inverse]
            assert is_symmetrically_sorted(symm_matrix_idxs[:, sorter][:, idx_inverse])
            symm_matrix_vals[mask_filter] = vector
            vector = symm_matrix_vals

    symm_matrix = torch.sparse.FloatTensor(idx, vector, symm_matrix.shape)
    torch.cuda.empty_cache()

    return symm_matrix


def dense2d_to_sparse_without_nonzero(tensor):
    x, y = tensor.shape
    nonzero = (tensor != 0).detach().cpu()
    x_idxs = torch.repeat_interleave(torch.arange(x), y)[nonzero.flatten()]
    y_idxs = torch.tile(torch.arange(y), [x])[nonzero.flatten()]
    indices = torch.stack((x_idxs, y_idxs)).to(tensor.device)
    values = tensor[nonzero]
    return torch.sparse.FloatTensor(indices, values, torch.Size((x, y)))


def create_symm_matrix_from_sparse_tril(tril):
    return (tril + tril.t()).to_dense()


def create_vec_from_symm_matrix(matrix):
    idx = torch.tril_indices(matrix.shape[0], matrix.shape[0])
    vector = matrix[idx[0], idx[1]]
    return vector


def get_sparse_eye_mat(num):
    i = torch.LongTensor([range(0, num), range(0, num)])
    val = torch.FloatTensor([1] * num)
    return torch.sparse.FloatTensor(i, val)


def unique_cat_recbole_interaction(inter, other, uid_field=None, iid_field=None, return_unique_counts=False):
    uid_field = uid_field or 'user_id'
    iid_field = iid_field or 'item_id'

    if isinstance(inter, dict):
        _inter = torch.stack((torch.as_tensor(inter[uid_field]), torch.as_tensor(inter[iid_field])))
    else:
        _inter = torch.as_tensor(inter)

    if isinstance(other, dict):
        _other = torch.stack((torch.as_tensor(other[uid_field]), torch.as_tensor(other[iid_field])))
    else:
        _other = torch.as_tensor(other)
    unique, counts = torch.cat((_inter, _other), dim=1).unique(dim=1, return_counts=True)
    new_inter = unique[:, counts == 1]

    if not return_unique_counts:
        return dict(zip([uid_field, iid_field], new_inter))
    else:
        return dict(zip([uid_field, iid_field], new_inter)), unique, counts


def np_unique_cat_recbole_interaction(inter, other, uid_field=None, iid_field=None, return_unique_counts=False):
    uid_field = uid_field or 'user_id'
    iid_field = iid_field or 'item_id'

    _inter = np.stack((inter[uid_field], inter[iid_field])) if isinstance(inter, dict) else inter
    _other = np.stack((other[uid_field], other[iid_field])) if isinstance(other, dict) else other

    unique, counts = np.unique(np.concatenate((_inter, _other), axis=1), axis=1, return_counts=True)
    new_inter = unique[:, counts == 1]

    if not return_unique_counts:
        return dict(zip([uid_field, iid_field], new_inter))
    else:
        return dict(zip([uid_field, iid_field], new_inter)), unique, counts


def remap_edges_recbole_ids(dataset, edges, field2id_token=True):
    mp_edges = []
    for i, _field in enumerate([dataset.uid_field, dataset.iid_field]):
        mp_edges.append([])
        for val in edges[i]:
            idx_val = val

            if field2id_token:
                if _field == dataset.iid_field:
                    idx_val = val - dataset.user_num

                mp_edges[-1].append(dataset.field2id_token[_field][idx_val])
            else:
                if _field == dataset.iid_field:
                    mp_val = dataset.field2token_id[_field][idx_val] + dataset.user_num
                else:
                    mp_val = dataset.field2token_id[_field][idx_val]

                mp_edges[-1].append(mp_val)
    return np.stack(mp_edges)


def rolling_window(input_array, size_kernel, stride, op="mean"):
    """Function to get rolling windows.
    https://stackoverflow.com/a/59781066

    Arguments:
        input_array {numpy.array} -- Input, by default it only works with depth equals to 1.
                                      It will be treated as a (height, width) image. If the input have (height, width, channel)
                                      dimensions, it will be rescaled to two-dimension (height, width)
        size_kernel {int} -- size of kernel to be applied. Usually 3,5,7. It means that a kernel of (size_kernel, size_kernel) will be applied
                             to the image.
        stride {int or tuple} -- horizontal and vertical displacement

    Returns:
        [list] -- A list with the resulting numpy.arrays
    """
    # Check right input dimension
    assert len(input_array.shape) in {1, 2}, \
        "input_array must have dimension 2 or 3. Yours have dimension {}".format(len(input_array))

    if input_array.shape == 3:
        input_array = input_array[:, :, 0]

    # Stride: horizontal and vertical displacement
    if isinstance(stride, int):
        sh, sw = stride, stride
    elif isinstance(stride, tuple):
        sh, sw = stride
    else:
        raise NotImplementedError("stride format not supported")

    # Input dimension (height, width)
    n_ah, n_aw = input_array.shape

    # Filter dimension (or window)
    if isinstance(size_kernel, int):
        n_h, n_w = size_kernel, size_kernel
    elif isinstance(size_kernel, tuple):
        n_h, n_w = size_kernel
    else:
        raise NotImplementedError("size_kernel format not supported")

    dim_out_h = int(np.floor((n_ah - n_h) / sh + 1))
    dim_out_w = int(np.floor((n_aw - n_w) / sw + 1))

    return _inner_rolling_window(input_array, dim_out_h, dim_out_w, n_h, n_w, sh, sw, op)


@numba.jit(nopython=True, parallel=True)
def _inner_rolling_window(input_array, dim_out_h, dim_out_w, n_h, n_w, sh, sw, op):
    # List to save output arrays
    rolled_array = np.zeros((dim_out_h, dim_out_w), dtype=np.float32)

    other_args = [dim_out_w, n_h, n_w, sh, sw]

    # Initialize row position
    for i in numba.prange(dim_out_h):
        start_row = sh * i
        _inner_inner_rolling_window(input_array, rolled_array, i, start_row, op, other_args)

    return rolled_array


@numba.jit(nopython=True)
def _inner_inner_rolling_window(input_array, rolled_array, i, start_row, op, other_args):
    dim_out_w, n_h, n_w, sh, sw = other_args
    for j in numba.prange(dim_out_w):
        start_col = sw * j

        # Get one window
        sub_array = input_array[start_row:(start_row + n_h), start_col:(start_col + n_w)]

        if op == "mean":
            agg_window = np.mean(sub_array)
        elif op == "sum":
            agg_window = np.sum(sub_array)
        else:
            raise TypeError("op should be one in ['sum', 'mean']")

        # Append sub_array
        rolled_array[i, j] = agg_window


def damerau_levenshtein_distance(s1, s2):
    s1 = [s1] if np.ndim(s1) == 1 else s1
    s2 = [s2] if np.ndim(s2) == 1 else s2

    out = np.zeros((len(s1, )), dtype=int)
    for i, (_s1, _s2) in enumerate(zip(s1, s2)):
        try:
            numb_s1, numb_s2 = numba.typed.List(_s1), numba.typed.List(_s2)
        except TypeError:  # python < 3.8
            numb_s1, numb_s2 = numba.typed.List(), numba.typed.List()
            for el in _s1:
                numb_s1.append(el)
            for el in _s2:
                numb_s2.append(el)

        out[i] = _damerau_levenshtein_distance(numb_s1, numb_s2)

    return out.item() if out.shape == (1,) else out


@numba.jit(nopython=True)
def _damerau_levenshtein_distance(s1, s2):
    """
    Copyright (c) 2015, James Turk
    https://github.com/jamesturk/jellyfish/blob/main/jellyfish/_jellyfish.py
    """

    # _check_type(s1)
    # _check_type(s2)

    len1 = len(s1)
    len2 = len(s2)
    infinite = len1 + len2

    # character array
    da = {}

    # distance matrix
    score = [[0] * (len2 + 2) for _ in range(len1 + 2)]

    score[0][0] = infinite
    for i in range(0, len1 + 1):
        score[i + 1][0] = infinite
        score[i + 1][1] = i
    for i in range(0, len2 + 1):
        score[0][i + 1] = infinite
        score[1][i + 1] = i

    for i in range(1, len1 + 1):
        db = 0
        for j in range(1, len2 + 1):
            i1 = da[s2[j - 1]] if s2[j - 1] in da else 0
            j1 = db
            cost = 1
            if s1[i - 1] == s2[j - 1]:
                cost = 0
                db = j

            score[i + 1][j + 1] = min(
                score[i][j] + cost,
                score[i + 1][j] + 1,
                score[i][j + 1] + 1,
                score[i1][j1] + (i - i1 - 1) + 1 + (j - j1 - 1),
            )
        da[s1[i - 1]] = i

    return score[len1 + 1][len2 + 1]


def copytree(src, dst, symlinks=False, ignore=None):
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    if ignore:
        excl = ignore(src, lst)
        lst = [x for x in lst if x not in excl]
    for item in lst:
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if symlinks and os.path.islink(s):
            if os.path.lexists(d):
                os.remove(d)
            os.symlink(os.readlink(s), d)
            try:
                st = os.lstat(s)
                mode = stat.S_IMODE(st.st_mode)
                os.lchmod(d, mode)
            except:
                pass  # lchmod not available
        elif os.path.isdir(s):
            copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def read_recbole_config_skip_errors(filepath, config):
    with open(filepath, 'r') as yaml_path:
        def construct_undefined(self, node):
            if isinstance(node, yaml.nodes.ScalarNode):
                value = self.construct_scalar(node)
            elif isinstance(node, yaml.nodes.SequenceNode):
                value = self.construct_sequence(node)
            elif isinstance(node, yaml.nodes.MappingNode):
                value = self.construct_mapping(node)
            else:
                assert False, f"unexpected node: {node!r}"
            return {node.__str__(): value}

        config.yaml_loader.add_constructor(None, construct_undefined)
        return yaml.load(yaml_path.read(), Loader=config.yaml_loader)



import os
import pickle
import pprint
import argparse

import tqdm
import pandas as pd

import gnnuers.data.utils as data_utils


if __name__ == "__main__":
    """
    # Windows
    python -m gnnuers.data.preprocess_dataset --train_split 0.7 --test_split 0.2 --validation_split 0.1 --split_type per_user --user_field user_id:token --item_field item_id:token --time_field timestamp:float --in_filepath dataset\lastfm-1k\lastfm-1k.inter --user_filepath dataset\lastfm-1k\lastfm-1k.user --dataset_name lastfm-1k_n-1 --min_interactions 20 --out_folderpath lastfm-1k_n-1\
    # Linux
    python3.9 -m gnnuers.data.preprocess_dataset --train_split 0.7 --test_split 0.2 --validation_split 0.1 --split_type per_user --user_field user_id:token --item_field item_id:token --time_field timestamp:float --in_filepath dataset/lastfm-1k/lastfm-1k.inter --user_filepath dataset/lastfm-1k/lastfm-1k.user --min_interactions 20 --dataset_name lastfm-1k_n-1 --out_folderpath lastfm-1k_n-1/

    # Train to Test (Linux)
    python3.9 -m gnnuers.data.preprocess_dataset --train_split 0.7 --test_split 0.2 --validation_split 0.1 --split_type per_user --user_field user_id:token --item_field item_id:token --time_field timestamp:float --in_filepath dataset/lastfm-1k/lastfm-1k.inter --user_filepath dataset/lastfm-1k/lastfm-1k.user --min_interactions 20 --dataset_name lastfm-1k_n-1 --out_folderpath lastfm-1k_n-1/ --train_to_test 1
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--in_filepath', required=True)
    parser.add_argument('--user_filepath', required=True)
    parser.add_argument('--out_folderpath', required=True)
    parser.add_argument('--dataset_name', required=True)
    parser.add_argument('--split_type', choices=['random', 'per_user'], default='per_user')
    parser.add_argument('--train_split', default=0.7, type=float)
    parser.add_argument('--train_to_test', default=0, type=int,
                        help="it moves the last N train interactions to test")
    parser.add_argument('--test_split', default=0.2, type=float)
    parser.add_argument('--validation_split', default=0.1, type=float)
    parser.add_argument('--user_field', default='user_id')
    parser.add_argument('--item_field', default='item_id')
    parser.add_argument('--time_field', default='timestamp')
    parser.add_argument('--random_state', type=int, default=120)
    parser.add_argument('--min_interactions', type=int, default=0)
    parser.add_argument('--add_token', action='store_true', help='add `token` or `float` to header')

    args = parser.parse_args()
    if args.train_to_test > 0:
        args.train_split = (args.train_split, args.train_to_test)

    df = pd.read_csv(args.in_filepath, sep='\t')
    user_df = pd.read_csv(args.user_filepath, sep='\t')

    pprint.pprint(vars(args))

    os.makedirs(args.out_folderpath, exist_ok=True)

    if args.min_interactions > 0:
        df = data_utils.filter_min_interactions(df, by=args.user_field, min_interactions=args.min_interactions)
        user_df = user_df[user_df[args.user_field].isin(df[args.user_field])]
        print(df.describe())
        print(df.apply(pd.unique, axis=0).apply(len))

    if args.add_token:
        if not 'token' in args.user_field:
            data_utils.add_token(df, user_df, args)
        else:
            print("`token` already present in headers")

    df.to_csv(os.path.join(args.out_folderpath, f"{args.dataset_name}.inter"), index=None, sep='\t')
    user_df.to_csv(os.path.join(args.out_folderpath, f"{args.dataset_name}.user"), index=None, sep='\t')

    if args.split_type == "per_user":
        print("> Splitting per user")
        train, val, test = data_utils.split_data_per_user(df,
                                                          train_split=args.train_split,
                                                          test_split=args.test_split,
                                                          validation_split=args.validation_split,
                                                          user_field=args.user_field,
                                                          time_field=args.time_field)
    elif args.split_type == "random":
        print("> Splitting randomly")
        train, val, test = data_utils.random_split(df,
                                                   train_split=args.train_split,
                                                   test_split=args.test_split,
                                                   validation_split=args.validation_split,
                                                   random_state=args.random_state)
    else:
        raise NotImplementedError(f"`split_type` = `{args.split_type}` not supported")

    remove_token = 'token' in args.user_field

    for data, data_name in zip([train, val, test], ["train", "validation", "test"]):
        with open(os.path.join(args.out_folderpath, f'{args.dataset_name}.{data_name}'), 'wb') as f:
            cols = [args.user_field, args.item_field]
            out_data = dict(zip(map(lambda x: x.split(':')[0] if remove_token else x, cols), data[cols].values.T))
            pickle.dump(out_data, f)



import torch
import torch.nn as nn
import torch.nn.functional as F

from recbole.model.general_recommender import NGCF as Recbole_NGCF


class NGCF(Recbole_NGCF):
    def __init__(self, config, dataset):
        super(NGCF, self).__init__(config, dataset)
        # NOT IN ORIGINAL CODE, ADDED TO PREVENT DROPOUT LAYER TO MODIFY VALUES EVEN IN EVAL MODE
        self.emb_dropout = nn.Dropout(self.message_dropout)

    def forward(self):

        A_hat = self.sparse_dropout(self.norm_adj_matrix) if self.node_dropout != 0 else self.norm_adj_matrix
        all_embeddings = self.get_ego_embeddings()
        embeddings_list = [all_embeddings]
        for gnn in self.GNNlayers:
            all_embeddings = gnn(A_hat, self.eye_matrix, all_embeddings)
            all_embeddings = nn.LeakyReLU(negative_slope=0.2)(all_embeddings)
            all_embeddings = self.emb_dropout(all_embeddings)
            all_embeddings = F.normalize(all_embeddings, p=2, dim=1)
            embeddings_list += [all_embeddings]  # storage output embedding of each layer
        ngcf_all_embeddings = torch.cat(embeddings_list, dim=1)

        user_all_embeddings, item_all_embeddings = torch.split(ngcf_all_embeddings, [self.n_users, self.n_items])

        return user_all_embeddings, item_all_embeddings



import os
import yaml
import shutil
import pickle
import inspect
import argparse
import logging

import numpy as np
import pandas as pd
from recbole.config import Config
from recbole.trainer import HyperTuning
from recbole.data import data_preparation
from recbole.utils import init_logger, get_model, get_trainer, init_seed, set_color, get_local_time

from gnnuers.models.ngcf import NGCF
from gnnuers.data import Dataset, PerturbedDataset
from gnnuers.explain import execute_explanation
import gnnuers.utils as utils


def training(_config, saved=True, model_file=None, hyper=False, perturbed_dataset=None):
    logger = logging.getLogger() if not hyper else None

    if model_file is not None and perturbed_dataset is None:
        _config, _model, _dataset, train_data, valid_data, test_data = utils.load_data_and_model(
            model_file,
            perturbed_dataset=perturbed_dataset
        )
    else:
        # dataset filtering
        _dataset = perturbed_dataset if perturbed_dataset is not None else Dataset(_config)

        # dataset splitting
        train_data, valid_data, test_data = data_preparation(_config, _dataset)

        # model loading and initialization
        if _config['model'].upper() == 'ngcf':
            _model = NGCF()(_config, train_data.dataset).to(_config['device'])
        else:
            _model = get_model(_config['model'])(_config, train_data.dataset).to(_config['device'])

        if not hyper:
            logger.info(_config)
            logger.info(_dataset)
            logger.info(_model)

    # trainer loading and initialization
    trainer = get_trainer(_config['MODEL_TYPE'], _config['model'])(_config, _model)
    if perturbed_dataset is not None:
        explanations_path = perturbed_dataset.explanations_path
        perturbed_suffix = "_PERTURBED"
        split_saved_file = os.path.basename(trainer.saved_model_file).split('-')
        perturbed_model_path = os.path.join(
            explanations_path,
            '-'.join(
                split_saved_file[:1] + [_dataset.dataset_name.upper()] + split_saved_file[1:]
            ).replace('.pth', '') + perturbed_suffix + '.pth'
        )

        resume_perturbed_training = False
        for f in os.scandir(explanations_path):
            if _config['model'].lower() in f.name.lower() and \
               _config['dataset'].lower() in f.name.lower() and \
               perturbed_suffix in f.name:
                perturbed_model_path = f.path
                resume_perturbed_training = True
                break

        trainer.saved_model_file = perturbed_model_path
        if resume_perturbed_training:
            trainer.resume_checkpoint(perturbed_model_path)
    elif model_file is not None:
        trainer.resume_checkpoint(model_file)
    else:
        split_saved_file = os.path.basename(trainer.saved_model_file).split('-')
        trainer.saved_model_file = os.path.join(
            os.path.dirname(trainer.saved_model_file),
            '-'.join(split_saved_file[:1] + [_dataset.dataset_name.upper()] + split_saved_file[1:])
        )

    # model training
    best_valid_score, best_valid_result = trainer.fit(
        train_data,
        valid_data,
        saved=saved,
        show_progress=_config['show_progress'] and not hyper,
        verbose=not hyper
    )

    # model evaluation
    test_result = trainer.evaluate(
        test_data,
        load_best_model=saved,
        show_progress=_config['show_progress'] and not hyper
    )

    if not hyper:
        logger.info(set_color('best valid ', 'yellow') + f': {best_valid_result}')
        logger.info(set_color('test result', 'yellow') + f': {test_result}')

    return {
        'best_valid_score': best_valid_score,
        'valid_score_bigger': _config['valid_metric_bigger'],
        'best_valid_result': best_valid_result,
        'test_result': test_result
    }


def recbole_hyper(base_config, params_file, config_file_list, saved=True):
    def objective_function(c_dict, c_file_list):
        config = Config(
            model=base_config['model'],
            dataset=base_config['dataset'],
            config_file_list=c_file_list,
            config_dict=c_dict
        )
        config['data_path'] = os.path.join(base_config.file_config_dict['data_path'], base_config.dataset)
        init_seed(base_config['seed'], config['reproducibility'])
        logging.basicConfig(level=logging.ERROR)

        return training(config, saved=False, hyper=True)

    hp = HyperTuning(
        objective_function, algo='exhaustive', params_file=params_file, fixed_config_file_list=config_file_list
    )
    hp.run()

    output_path = os.path.join(base_config['checkpoint_dir'], 'hyper', base_config['dataset'], base_config['model'])
    os.makedirs(output_path, exist_ok=True)
    output_file = os.path.join(output_path, f"{get_local_time()}.txt")

    hp.export_result(output_file=output_file)
    print('best params: ', hp.best_params)
    print('best result: ')
    print(hp.params2result[hp.params2str(hp.best_params)])

    with open(output_file, 'r+') as f:
        content = f.read()
        f.seek(0, 0)
        f.write(
            'Best Params and Results\n' +
            str(hp.best_params).rstrip('\r\n') + '\n' +
            str(hp.params2result[hp.params2str(hp.best_params)]) + '\n\n' + content
        )


def main(model=None, dataset=None, config_file_list=None, config_dict=None, saved=True, seed=None, hyper_params_file=None):
    r""" A fast running api, which includes the complete process of
    training and testing a model on a specified dataset
    Args:
        model (str, optional): Model name. Defaults to ``None``.
        dataset (str, optional): Dataset name. Defaults to ``None``.
        config_file_list (list, optional): Config files used to modify experiment parameters. Defaults to ``None``.
        config_dict (dict, optional): Parameters dictionary used to modify experiment parameters. Defaults to ``None``.
        saved (bool, optional): Whether to save the model. Defaults to ``True``.
    """
    # configurations initialization
    config = Config(model=model, dataset=dataset, config_file_list=config_file_list, config_dict=config_dict)
    config['data_path'] = os.path.join(config.file_config_dict['data_path'], config.dataset)
    seed = seed or config['seed']
    init_seed(seed, config['reproducibility'])
    import torch; torch.use_deterministic_algorithms(True)

    # logger initialization
    init_logger(config)
    logger = logging.getLogger()

    # if args.run == 'evaluate_perturbed' or args.run == 'graph_stats':
    #     orig_config, orig_model, orig_dataset, orig_train_data, orig_valid_data, orig_test_data = \
    #         utils.load_data_and_model(args.original_model_file, args.explainer_config_file)

    if args.use_perturbed_graph:
        perturbed_dataset = PerturbedDataset(config, args.explanations_path, args.best_exp)
        if args.run == 'train':
            training(config, saved=saved, model_file=args.model_file, perturbed_dataset=perturbed_dataset)
            # elif args.run == 'explain':
            #     runner(*explain_args)
            # elif args.run == 'evaluate_perturbed':
            #     logger.info("EVALUATE PERTURBED MODEL")
            #     _, pert_model, pert_dataset, _, _, _ = utils.load_data_and_model(args.model_file,
            #                                                                      args.explainer_config_file)
            #     runner(
            #         orig_config,
            #         orig_model,
            #         pert_model,
            #         orig_dataset,
            #         pert_dataset,
            #         orig_train_data,
            #         orig_test_data,
            #         topk=args.topk,
            #         perturbed_model_file=os.path.splitext(os.path.basename(args.model_file))[0]
            #     )
            # elif args.run == 'graph_stats':
            #     pert_config, _, _, pert_train_data, _, _ = utils.load_data_and_model(args.model_file,
            #                                                                          args.explainer_config_file)
            #     runner(
            #         pert_config,
            #         orig_train_data,
            #         orig_valid_data,
            #         orig_test_data,
            #         pert_train_data,
            #         args.original_model_file,
            #         sens_attr,
            #         c_id,
            #         *args.best_exp
            #     )
    else:
        # dataset = create_dataset(config)
        # logger.info(dataset)

        if args.run == 'train':
            training(config, saved=saved, model_file=args.model_file)
        elif args.run == 'explain':
            execute_explanation(args.model_file, *explain_args)
        elif args.run == 'recbole_hyper':
            config['seed'] = seed
            recbole_hyper(config, hyper_params_file, config_file_list, saved=saved)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    perturbed_train_group = parser.add_argument_group(
        "perturbed_train",
        "All the arguments related to training with augmented data"
    )
    explain_group = parser.add_argument_group(
        "explain",
        "All the arguments related to create explanations"
    )
    recbole_hyper_group = parser.add_argument_group(
        "recole_hyper",
        "All the arguments related to run the hyperparameter optimization on the recbole models for training"
    )

    parser.add_argument('--run', default='train', choices=['train', 'explain', 'recbole_hyper'], required=True)
    parser.add_argument('--seed', default=None, type=int)
    parser.add_argument('--model', default='GCMC')
    parser.add_argument('--dataset', default='ml-100k')
    parser.add_argument('--config_file_list', nargs='+', default=None)
    parser.add_argument('--config_dict', default=None)
    parser.add_argument('--model_file', default=None)
    explain_group.add_argument('--base_explainer_config_file', default=os.path.join("config", "base_explainer.yaml"))
    explain_group.add_argument('--explainer_config_file', default=os.path.join("config", "explainer.yaml"))
    # explain_group.add_argument('--load', action='store_true')
    explain_group.add_argument('--explain_config_id', default=-1)
    explain_group.add_argument('--verbose', action='store_true')
    explain_group.add_argument('--wandb_online', action='store_true')
    explain_group.add_argument('--hyper_optimize', action='store_true')
    explain_group.add_argument('--overwrite', action='store_true')
    perturbed_train_group.add_argument('--use_perturbed_graph', action='store_true')
    perturbed_train_group.add_argument('--best_exp', nargs="*",
                                       help="one of ['fairest', 'fairest_before_exp', 'fixed_exp'] with "
                                            "the chosen exp number for the last two types")
    perturbed_train_group.add_argument('--explanations_path', default=None)
    recbole_hyper_group.add_argument('--params_file', default=None)

    args, unk_args = parser.parse_known_args()
    print(args)

    unk_args[::2] = map(lambda s: s.replace('-', ''), unk_args[::2])
    unk_args = dict(zip(unk_args[::2], unk_args[1::2]))
    print("Unknown args", unk_args)

    if args.hyper_optimize and not args.verbose:
        from tqdm import tqdm
        from functools import partialmethod

        tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)

    if args.use_perturbed_graph:
        if args.explainer_config_file:
            # TODO: check that the path has the experiments folder setup
            pass

    args.wandb_online = {False: "offline", True: "online"}[args.wandb_online]
    explain_args = [
        args.base_explainer_config_file,
        args.explainer_config_file,
        args.explain_config_id,
        args.verbose,
        args.wandb_online,
        unk_args,
        args.hyper_optimize,
        args.overwrite
    ]

    if args.run not in ['train', 'explain', 'recbole_hyper']:
        raise NotImplementedError(f"The run `{args.run}` is not supported.")

    main(
        args.model,
        args.dataset,
        args.config_file_list,
        args.config_dict,
        seed=args.seed,
        hyper_params_file=args.params_file
    )
